"""LimitlessLED.

http://www.limitlessled.com
"""

from collections import namedtuple
import logging

_LOGGER = logging.getLogger(__name__)


# Various constants.
MIN_WAIT = 0.1
REPS = 3
MAX_GROUPS = 4


# Color tuple.
Color = namedtuple("Color", "R G B")


class LimitlessLED(object):
    """Represents a LimitlessLED installation."""

    def __init__(self):
        """Initialize."""
        self._groups = {}

    def group(self, name):
        """Fetch a named group.

        :param name: Name of group.
        :returns: Group.
        """
        return self._groups[name]

    def add_bridge(self, bridge):
        """Add bridge groups.

        :param bridge: Add groups from this bridge.
        """
        for group in bridge.groups:
            self._groups[group.name] = group



"""Command sets for wifi bridge version 6."""

import math

from limitlessled.group.commands import Command, CommandSet
from limitlessled.group.dimmer import DIMMER
from limitlessled.group.rgbcct import RGBCCT
from limitlessled.group.rgbw import BRIDGE_LED, RGBW
from limitlessled.group.rgbww import RGBWW
from limitlessled.group.white import WHITE
from limitlessled.group.wrgb import WRGB


class CommandV6(Command):
    """Represents a single v6 command to be sent to the bridge."""

    PASSWORD_BYTE1 = 0x00
    PASSWORD_BYTE2 = 0x00

    TYPE_CONTROL = 0x31
    TYPE_LINK = 0x3D
    TYPE_UNLINK = 0x3E

    def __init__(
        self,
        cmd_1,
        cmd_2,
        remote_style,
        group_number,
        select=False,
        select_command=None,
        command_type=0x31,
    ):
        """
        Initialize command.
        :param cmd_1: The first part of the command.
        :param cmd_1: The second part of the command.
        :param remote_style: The remote style of the led.
        :param group_number: Group number (1-4).
        :param select: If command requires selection.
        :param select_command: Selection command bytes.
        :param command_type: Whether the command is for control, link, or unlink
        """
        super().__init__(cmd_1, cmd_2, group_number, select, select_command)
        self._remote_style = remote_style
        self.type = command_type

    def get_bytes(self, bridge):
        """
        Gets the full command as bytes.
        :param bridge: The bridge, to which the command should be sent.
        """
        if not bridge.is_ready:
            raise Exception("The bridge has to be ready to construct command.")

        wb1 = bridge.wb1
        wb2 = bridge.wb2
        sn = bridge.sn

        preamble = [0x80, 0x00, 0x00, 0x00, 0x11, wb1, wb2, 0x00, sn, 0x00]
        cmd = [
            self.type,
            self.PASSWORD_BYTE1,
            self.PASSWORD_BYTE2,
            self._remote_style,
            self._cmd_1,
            self._cmd_2,
            self._cmd_2,
            self._cmd_2,
            self._cmd_2,
        ]
        zone_selector = [self._group_number, 0x00]
        checksum = sum(cmd + zone_selector) & 0xFF

        return bytearray(preamble + cmd + zone_selector + [checksum])


class CommandSetV6(CommandSet):
    """Base command set for wifi bridge v6."""

    SUPPORTED_VERSIONS = [6]
    MAX_HUE = 0xFF
    MAX_SATURATION = 0x64
    MAX_BRIGHTNESS = 0x64
    MAX_TEMPERATURE = 0x64
    MAX_ZONE_NUMBER = 4

    def __init__(
        self,
        group_number,
        remote_style,
        brightness_steps=None,
        hue_steps=None,
        saturation_steps=None,
        temperature_steps=None,
    ):
        """
        Initialize the command set.
        :param group_number: The group number.
        :param remote_style: The remote style of the device to control.
        :param brightness_steps: The number of brightness steps.
        :param hue_steps: The number of color steps.
        :param saturation_steps: The number of saturation steps
        :param temperature_steps: The number of temperature steps.
        """
        group_number = min(max(group_number, 0), self.MAX_ZONE_NUMBER)
        brightness_steps = brightness_steps or self.MAX_BRIGHTNESS + 1
        hue_steps = hue_steps or self.MAX_HUE + 1
        saturation_steps = saturation_steps or self.MAX_SATURATION + 1
        temperature_steps = temperature_steps or self.MAX_TEMPERATURE + 1
        super().__init__(
            group_number,
            brightness_steps,
            hue_steps=hue_steps,
            saturation_steps=saturation_steps,
            temperature_steps=temperature_steps,
        )
        self._remote_style = remote_style

    def convert_brightness(self, brightness):
        """
        Convert the brightness from decimal percent (0.0-1.0)
        to byte representation for use in commands.
        :param brightness: The brightness from in decimal percent (0.0-1.0).
        :return: The brightness in byte representation.
        """
        return math.ceil(brightness * self.MAX_BRIGHTNESS)

    def convert_saturation(self, saturation):
        """
        Convert the saturation from decimal percent (0.0-1.0)
        to byte representation for use in commands.
        :param saturation: The saturation from in decimal percent (0.0-1.0).
        1.0 is the maximum saturation where no white leds will be on. 0.0 is no
        saturation.
        :return: The saturation in byte representation.
        """

        saturation_inverted = 1 - saturation
        return math.ceil(saturation_inverted * self.MAX_SATURATION)

    def convert_temperature(self, temperature):
        """
        Convert the temperature from decimal percent (0.0-1.0)
        to byte representation for use in commands.
        :param temperature: The temperature from in decimal percent (0.0-1.0).
        :return: The temperature in byte representation.
        """
        return math.ceil(temperature * self.MAX_TEMPERATURE)

    def convert_hue(self, hue, legacy_color_wheel=False):
        """
        Converts the hue from HSV color circle to the LimitlessLED color wheel.
        :param hue: The hue in decimal percent (0.0-1.0).
        :param legacy_color_wheel: Whether or not use the old color wheel.
        :return: The hue regarding the LimitlessLED color wheel.
        """
        hue = math.ceil(hue * self.MAX_HUE)
        if legacy_color_wheel:
            hue = (176 - hue) % (self.MAX_HUE + 1)
            hue = (self.MAX_HUE - hue - 0x37) % (self.MAX_HUE + 1)
        else:
            hue += 10  # The color wheel for RGBWW bulbs seems to be shifted

        return hue % (self.MAX_HUE + 1)

    def _build_command(self, cmd_1, cmd_2, select=False, select_command=None, command_type=0x31):
        """
        Constructs the complete command.
        :param cmd_1: Light command 1.
        :param cmd_2: Light command 2.
        :param command_type: Whether the command is for control, link, or unlink
        :return: The complete command.
        """

        return CommandV6(
            cmd_1,
            cmd_2,
            self._remote_style,
            self._group_number,
            select,
            select_command,
            command_type,
        )


class CommandSetBridgeLightV6(CommandSetV6):
    """Command set for bridge light of wifi bridge v6."""

    SUPPORTED_LED_TYPES = [BRIDGE_LED]
    REMOTE_STYLE = 0x00

    def __init__(self, group_number):
        """
        Initializes the command set.
        :param group_number: The group number.
        """
        super().__init__(group_number, self.REMOTE_STYLE)

    def on(self):
        """
        Build command for turning the led on.
        :return: The command.
        """
        return self._build_command(0x03, 0x03)

    def off(self):
        """
        Build command for turning the led off.
        :return: The command.
        """
        return self._build_command(0x03, 0x04)

    def white(self):
        """
        Build command for turning the led into white mode.
        :return: The command.
        """
        return self._build_command(0x03, 0x05)

    def hue(self, hue):
        """
        Build command for setting the hue of the led.
        :param hue: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x01, self.convert_hue(hue))

    def brightness(self, brightness):
        """
        Build command for setting the brightness of the led.
        :param brightness: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x02, self.convert_brightness(brightness))


class CommandSetWhiteV6(CommandSetV6):
    """Command set for white led light connected to wifi bridge v6."""

    SUPPORTED_LED_TYPES = [WHITE]
    REMOTE_STYLE = 0x01

    def __init__(self, group_number):
        """
        Initializes the command set.
        :param group_number: The group number.
        """
        super().__init__(group_number, self.REMOTE_STYLE)

    def on(self):
        """
        Build command for turning the led on.
        :return: The command.
        """
        return self._build_command(0x01, 0x07)

    def off(self):
        """
        Build command for turning the led off.
        :return: The command.
        """
        return self._build_command(0x01, 0x08)

    def link(self):
        """
        Build command for linking new lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_LINK)

    def unlink(self):
        """
        Build command for unlinking linked lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_UNLINK)

    def night_light(self):
        """
        Build command for turning the led into night light mode.
        :return: The command.
        """
        return self._build_command(0x01, 0x06)

    def dimmer(self):
        """
        Build command for setting the brightness one step dimmer.
        :return: The command.
        """
        return self._build_command(0x01, 0x02, select=True, select_command=self.on())

    def brighter(self):
        """
        Build command for setting the brightness one step brighter.
        :return: The command.
        """
        return self._build_command(0x01, 0x01, select=True, select_command=self.on())

    def cooler(self):
        """
        Build command for setting the temperature one step cooler.
        :return: The command.
        """
        return self._build_command(0x01, 0x04, select=True, select_command=self.on())

    def warmer(self):
        """
        Build command for setting the temperature one step warmer.
        :return: The command.
        """
        return self._build_command(0x01, 0x03, select=True, select_command=self.on())


class CommandSetDimmerV6(CommandSetV6):
    """Command set for Dimmer LED dimmer (1CH MiLight dimmer) connected to wifi bridge v6."""

    SUPPORTED_LED_TYPES = [DIMMER]
    REMOTE_STYLE = 0x03

    def __init__(self, group_number):
        """
        Initializes the command set.
        :param group_number: The group number.
        """
        super().__init__(group_number, self.REMOTE_STYLE)

    def on(self):
        """
        Build command for turning the dimmer on.
        :return: The command.
        """
        return self._build_command(0x04, 0x03)

    def off(self):
        """
        Build command for turning the dimmer off.
        :return: The command.
        """
        return self._build_command(0x04, 0x04)

    def link(self):
        """
        Build command for linking new lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_LINK)

    def unlink(self):
        """
        Build command for unlinking linked lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_UNLINK)

    def night_light(self):
        """
        Build command for turning the dimmer into night light mode.
        :return: The command.
        """
        return self._build_command(0x04, 0x02)

    def brightness(self, brightness):
        """
        Build command for setting the brightness of the controller.
        :param brightness: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x01, self.convert_brightness(brightness))


class CommandSetRgbwV6(CommandSetV6):
    """Command set for RGBW led light connected to wifi bridge v6."""

    SUPPORTED_LED_TYPES = [RGBW]
    REMOTE_STYLE = 0x07

    def __init__(self, group_number):
        """
        Initializes the command set.
        :param group_number: The group number.
        """
        super().__init__(group_number, self.REMOTE_STYLE)

    def on(self):
        """
        Build command for turning the led on.
        :return: The command.
        """
        return self._build_command(0x03, 0x01)

    def off(self):
        """
        Build command for turning the led off.
        :return: The command.
        """
        return self._build_command(0x03, 0x02)

    def link(self):
        """
        Build command for linking new lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_LINK)

    def unlink(self):
        """
        Build command for unlinking linked lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_UNLINK)

    def night_light(self):
        """
        Build command for turning the led into night light mode.
        :return: The command.
        """
        return self._build_command(0x03, 0x06)

    def white(self):
        """
        Build command for turning the led into white mode.
        :return: The command.
        """
        return self._build_command(0x03, 0x05)

    def hue(self, hue):
        """
        Build command for setting the hue of the led.
        :param hue: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x01, self.convert_hue(hue, True))

    def brightness(self, brightness):
        """
        Build command for setting the brightness of the led.
        :param brightness: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x02, self.convert_brightness(brightness))


class CommandSetWrgbV6(CommandSetV6):
    """Command set for WRGB led light connected to wifi bridge v6."""

    SUPPORTED_LED_TYPES = [WRGB]
    REMOTE_STYLE = 0x06

    def __init__(self, group_number):
        """
        Initializes the command set.
        :param group_number: The group number.
        """
        super().__init__(group_number, self.REMOTE_STYLE)

    def on(self):
        """
        Build command for turning the led on.
        :return: The command.
        """
        return self._build_command(0x03, 0x01)

    def off(self):
        """
        Build command for turning the led off.
        :return: The command.
        """
        return self._build_command(0x03, 0x02)

    def link(self):
        """
        Build command for linking new lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_LINK)

    def unlink(self):
        """
        Build command for unlinking linked lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_UNLINK)

    def night_light(self):
        """
        Build command for turning the led into night light mode.
        :return: The command.
        """
        return self._build_command(0x03, 0x06)

    def white(self):
        """
        Build command for turning the led into white mode.
        :return: The command.
        """
        return self._build_command(0x03, 0x05)

    def white_up(self):
        """
        Build command for turning the white channel up
        :return: The command.
        """
        return self._build_command(0x03, 0x03)

    def white_down(self):
        """
        Build command for turning the white channel doen
        :return: The command.
        """
        return self._build_command(0x03, 0x04)

    def red_up(self):
        """
        Build command for turning the red channel up
        :return: The command.
        """
        return self._build_command(0x03, 0x05)

    def red_down(self):
        """
        Build command for turning the red channel doen
        :return: The command.
        """
        return self._build_command(0x03, 0x06)

    def green_up(self):
        """
        Build command for turning the green channel up
        :return: The command.
        """
        return self._build_command(0x03, 0x07)

    def green_down(self):
        """
        Build command for turning the green channel doen
        :return: The command.
        """
        return self._build_command(0x03, 0x08)

    def blue_up(self):
        """
        Build command for turning the blue channel up
        :return: The command.
        """
        return self._build_command(0x03, 0x09)

    def blue_down(self):
        """
        Build command for turning the blue channel doen
        :return: The command.
        """
        return self._build_command(0x03, 0x0A)

    def hue(self, hue):
        """
        Build command for setting the hue of the led.
        :param hue: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x01, self.convert_hue(hue, True))

    def brightness(self, brightness):
        """
        Build command for setting the brightness of the led.
        :param brightness: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x02, self.convert_brightness(brightness))


class CommandSetRgbwwV6(CommandSetV6):
    """Command set for RGBWW led light connected to wifi bridge v6."""

    SUPPORTED_LED_TYPES = [RGBWW]
    REMOTE_STYLE = 0x08

    def __init__(self, group_number):
        """
        Initializes the command set.
        :param group_number: The group number.
        """
        super().__init__(group_number, self.REMOTE_STYLE)

    def on(self):
        """
        Build command for turning the led on.
        :return: The command.
        """
        return self._build_command(0x04, 0x01)

    def off(self):
        """
        Build command for turning the led off.
        :return: The command.
        """
        return self._build_command(0x04, 0x02)

    def link(self):
        """
        Build command for linking new lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_LINK)

    def unlink(self):
        """
        Build command for unlinking linked lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_UNLINK)

    def night_light(self):
        """
        Build command for turning the led into night light mode.
        :return: The command.
        """
        return self._build_command(0x04, 0x05)

    def white(self, temperature=1):
        """
        Build command for turning the led into white mode.
        :param: The temperature to set.
        :return: The command.
        """
        return self.temperature(temperature)

    def hue(self, hue):
        """
        Build command for setting the hue of the led.
        :param hue: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x01, self.convert_hue(hue))

    def saturation(self, saturation):
        """
        Build command for setting the saturation of the led.
        :param saturation: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x02, self.convert_saturation(saturation))

    def brightness(self, brightness):
        """
        Build command for setting the brightness of the led.
        :param brightness: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x03, self.convert_brightness(brightness))

    def temperature(self, temperature):
        """
        Build command for setting the temperature of the led.
        :param temperature: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x05, self.convert_temperature(temperature))


class CommandSetRgbcctV6(CommandSetV6):
    """Command set for RGBCCT led light connected to wifi bridge v6."""

    SUPPORTED_LED_TYPES = [RGBCCT]
    REMOTE_STYLE = 0x0A
    MAX_ZONE_NUMBER = 8

    def __init__(self, group_number):
        """
        Initializes the command set.
        :param group_number: The group number.
        """
        super().__init__(group_number, self.REMOTE_STYLE)

    def on(self):
        """
        Build command for turning the led on.
        :return: The command.
        """
        return self._build_command(0x06, 0x01)

    def off(self):
        """
        Build command for turning the led off.
        :return: The command.
        """
        return self._build_command(0x06, 0x02)

    def link(self):
        """
        Build command for linking new lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_LINK)

    def unlink(self):
        """
        Build command for unlinking linked lights.
        :return: The command.
        """
        return self._build_command(0x00, 0x00, command_type=CommandV6.TYPE_UNLINK)

    def night_light(self):
        """
        Build command for turning the led into night light mode.
        :return: The command.
        """
        return self._build_command(0x06, 0x64)

    def white(self):
        """
        Build command for turning the led into white mode.
        :return: The command.
        """
        return self._build_command(0x06, 0x05)

    def hue(self, hue):
        """
        Build command for setting the hue of the led.
        :param hue: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x01, self.convert_hue(hue))

    def saturation(self, saturation):
        """
        Build command for setting the saturation of the led.
        :param saturation: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x03, self.convert_saturation(saturation))

    def brightness(self, brightness):
        """
        Build command for setting the brightness of the led.
        :param brightness: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x04, self.convert_brightness(brightness))

    def temperature(self, temperature):
        """
        Build command for setting the temperature of the led.
        :param temperature: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(0x02, self.convert_temperature(temperature))



from limitlessled import Color
from limitlessled.group import Group
from limitlessled.group.rgbww import RgbwwGroup
from limitlessled.util import rgb_to_hsv

RGBCCT = "rgbcct"
RGB_WHITE = Color(255, 255, 255)


class RgbcctGroup(RgbwwGroup):
    """RGBCCT LimitlessLED group."""

    def __init__(self, bridge, number, name):
        """Initialize RGBCCT group.

        :param bridge: Associated bridge.
        :param number: Group number (1-4).
        :param name: Group name.
        """
        Group.__init__(self, bridge, number, name, RGBCCT)
        self._saturation = 0
        self._hue = 0
        self._temperature = 0.5
        self._color = RGB_WHITE

    def white(self):
        """Set color to white."""
        self._color = RGB_WHITE
        cmd = self.command_set.white()
        self.send(cmd)

    @property
    def color(self):
        """Color property.

        :returns: Color.
        """
        return self._color

    @color.setter
    def color(self, color):
        """Set group color.

        Color is set on a best-effort basis.

        :param color: RGB color tuple.
        """
        self.hue, self.saturation, self.brightness = rgb_to_hsv(*[x / 255 for x in color])



"""Various preset pipelines."""

from limitlessled import Color
from limitlessled.pipeline import Pipeline

# Alarm (flash red).
ALARM = Pipeline().on().color(255, 0, 0).flash().repeat()


# Color loop (R->G->B).
COLORLOOP = (
    Pipeline()
    .on()
    .transition(10, color=Color(255, 0, 0))
    .transition(10, color=Color(0, 255, 0))
    .transition(10, color=Color(0, 0, 255))
    .repeat(stages=3)
)



"""RGBW LimitlessLED group."""

import math
import time

from limitlessled import Color, util
from limitlessled.group import Group, rate
from limitlessled.util import hue_of_color, saturation_of_color, steps, to_rgb

RGBWW = "rgbww"
RGB_WHITE = Color(255, 255, 255)


class RgbwwGroup(Group):
    """RGBWW LimitlessLED group."""

    def __init__(self, bridge, number, name):
        """Initialize RGBWW group.

        :param bridge: Associated bridge.
        :param number: Group number (1-4).
        :param name: Group name.
        """
        super().__init__(bridge, number, name, RGBWW)
        self._saturation = 0
        self._hue = 0
        self._temperature = 0.5
        self._color = RGB_WHITE

    @property
    def color(self):
        """Color property.

        :returns: Color.
        """
        return self._color

    @color.setter
    def color(self, color):
        """Set group color.

        Color is set on a best-effort basis.

        :param color: RGB color tuple.
        """
        self._color = color
        self.saturation = saturation_of_color(color)
        if self.saturation != 0:
            self.hue = hue_of_color(color)

    def white(self):
        """Set color to white."""
        self._color = RGB_WHITE
        cmd = self.command_set.white(self.temperature)
        self.send(cmd)

    def night_light(self):
        """Set night light mode."""
        cmd = self.command_set.night_light()
        self.send(cmd)

    @property
    def brightness(self):
        """Brightness property.

        :returns: Brightness.
        """
        return self._brightness

    @brightness.setter
    def brightness(self, brightness):
        """Set the group brightness.

        :param brightness: Brightness in decimal percent (0.0-1.0).
        """
        if brightness < 0 or brightness > 1:
            raise ValueError("Brightness must be a percentage represented as decimal 0-1.0")
        self._brightness = brightness
        cmd = self.command_set.brightness(brightness)
        self.send(cmd)

    @property
    def hue(self):
        """Hue property.

        :returns: Hue.
        """
        return self._hue

    @hue.setter
    def hue(self, hue):
        """Set the group hue.

        :param hue: Hue in decimal percent (0.0-1.0).
        """
        if hue < 0 or hue > 1:
            raise ValueError("Hue must be a percentage represented as decimal 0-1.0")
        self._hue = hue
        self._update_color()
        cmd = self.command_set.hue(hue)
        self.send(cmd)

    @property
    def saturation(self):
        """Saturation property.

        :returns: Saturation.
        """
        return self._saturation

    @saturation.setter
    def saturation(self, saturation):
        """Set the group saturation.

        :param saturation: Saturation in decimal percent (0.0-1.0).
        """
        if saturation < 0 or saturation > 1:
            raise ValueError("Saturation must be a percentage represented as decimal 0-1.0")
        self._saturation = saturation
        self._update_color()
        if saturation == 0:
            self.white()
        else:
            cmd = self.command_set.saturation(saturation)
            self.send(cmd)

    def _update_color(self):
        """Update the color property from hue and saturation values."""
        self._color = to_rgb(self.hue, self.saturation)

    @property
    def temperature(self):
        """Temperature property.

        :returns: Temperature (0.0-1.0)
        """
        return self._temperature

    @temperature.setter
    def temperature(self, temperature):
        """Set the temperature.

        :param temperature: Value to set (0.0-1.0).
        """
        if temperature < 0 or temperature > 1:
            raise ValueError("Temperature must be a percentage represented as decimal 0-1.0")
        self._temperature = temperature
        cmd = self.command_set.temperature(temperature)
        self.send(cmd)

    def transition(self, duration, color=None, brightness=None, temperature=None):
        """Transition wrapper.

        Short-circuit transition as necessary.

        :param duration: Time to transition.
        :param color: Transition to this color.
        :param brightness: Transition to this brightness.
        :param temperature: Transition to this temperature.
        """
        if color and temperature is not None:
            raise ValueError("Cannot transition to color and temperature simultaneously.")

        # Transition to white immediately.
        if color == RGB_WHITE:
            self.white()
        # Transition away from white immediately.
        elif self.color == RGB_WHITE and color is not None:
            self.color = color
        # Transition immediately if duration is zero.
        if duration == 0:
            if brightness is not None:
                self.brightness = brightness
            if color:
                self.color = color
            if temperature is not None:
                self.temperature = temperature
            return
        # Perform transition
        if color and color != self.color:
            self._transition(
                duration, brightness, hue=hue_of_color(color), saturation=saturation_of_color(color)
            )
        elif temperature != self.temperature:
            self._transition(duration, brightness, temperature=temperature)
        elif brightness != self.brightness:
            self._transition(duration, brightness)

    @rate(wait=0.025, reps=1)
    def _transition(self, duration, brightness, hue=None, saturation=None, temperature=None):
        """Transition.

        :param duration: Time to transition.
        :param brightness: Transition to this brightness.
        :param hue: Transition to this hue.
        :param saturation: Transition to this saturation.
        :param temperature: Transition to this temperature.
        """
        # Calculate brightness steps.
        b_steps = 0
        if brightness is not None:
            b_steps = steps(self.brightness, brightness, self.command_set.brightness_steps)
            b_start = self.brightness
        # Calculate hue steps.
        h_steps = 0
        if hue is not None:
            h_steps = steps(self.hue, hue, self.command_set.hue_steps)
            h_start = self.hue
        # Calculate saturation steps.
        s_steps = 0
        if saturation is not None:
            s_steps = steps(self.saturation, saturation, self.command_set.saturation_steps)
            s_start = self.saturation
        # Calculate temperature steps.
        t_steps = 0
        if temperature is not None:
            t_steps = steps(self.temperature, temperature, self.command_set.temperature_steps)
            t_start = self.temperature
        # Compute ideal step amount (at least one).
        total_steps = max(b_steps, h_steps, s_steps, t_steps, 1)
        total_commands = b_steps + h_steps + s_steps + t_steps
        # Calculate wait.
        wait = self._wait(duration, total_steps, total_commands)
        # Scale down steps if no wait time.
        if wait == 0:
            scaled_steps = self._scale_steps(
                duration, total_commands, b_steps, h_steps, s_steps, t_steps
            )
            b_steps, h_steps, s_steps, t_steps = scaled_steps
            total_steps = max(b_steps, h_steps, s_steps, t_steps, 1)
        # Perform transition.
        for i in range(total_steps):
            # Brightness.
            if b_steps > 0 and i % math.ceil(total_steps / b_steps) == 0:
                self.brightness = util.transition(i, total_steps, b_start, brightness)
            # Hue.
            if h_steps > 0 and i % math.ceil(total_steps / h_steps) == 0:
                self.hue = util.transition(i, total_steps, h_start, hue)
            # Saturation.
            if s_steps > 0 and i % math.ceil(total_steps / s_steps) == 0:
                self.saturation = util.transition(i, total_steps, s_start, saturation)
            # Temperature.
            if t_steps > 0 and i % math.ceil(total_steps / t_steps) == 0:
                self.temperature = util.transition(i, total_steps, t_start, temperature)

            # Wait.
            time.sleep(wait)



"""LimtlessLED command sets."""

import binascii


def command_set_factory(bridge, group_number, led_type):
    """
    Create command set for controlling a specific led group.
    :param bridge: The bridge the leds are connected to.
    :param group_number: The group number.
    :param led_type: The type of the leds.
    :return: The created command set.
    """
    from limitlessled.group.commands.legacy import CommandSetRgbwLegacy, CommandSetWhiteLegacy
    from limitlessled.group.commands.v6 import (
        CommandSetBridgeLightV6,
        CommandSetDimmerV6,
        CommandSetRgbcctV6,
        CommandSetRgbwV6,
        CommandSetRgbwwV6,
        CommandSetWhiteV6,
        CommandSetWrgbV6,
    )

    command_sets = [
        CommandSetWhiteLegacy,
        CommandSetRgbwLegacy,
        CommandSetBridgeLightV6,
        CommandSetWhiteV6,
        CommandSetDimmerV6,
        CommandSetRgbwV6,
        CommandSetRgbwwV6,
        CommandSetWrgbV6,
        CommandSetRgbcctV6,
    ]
    try:
        cls = next(
            cs
            for cs in command_sets
            if bridge.version in cs.SUPPORTED_VERSIONS and led_type in cs.SUPPORTED_LED_TYPES
        )
        return cls(group_number)
    except StopIteration:
        raise ValueError("There is no command set for specified bridge version and led type.")


class Command:
    """Base class for a single command to be sent to the bridge."""

    def __init__(self, cmd_1, cmd_2, group_number, select=False, select_command=None):
        """
        Initialize command.
        :param cmd_1: The first part of the command.
        :param cmd_2: The second part of the command.
        :param group_number: Group number (1-4).
        :param select: If command requires selection.
        :param select_command: Selection command bytes.
        """
        self._cmd_1 = cmd_1
        self._cmd_2 = cmd_2
        self._group_number = group_number
        self._select = select
        self._select_command = select_command

    @property
    def cmd_1(self):
        """The first part of the raw command."""
        return self._cmd_1

    @property
    def cmd_2(self):
        """The first part of the raw command."""
        return self._cmd_2

    @property
    def group_number(self):
        """The group number (1-4)."""
        return self._group_number

    @property
    def select(self):
        """If command requires selection."""
        return self._select

    @property
    def select_command(self):
        """Selection command bytes."""
        return self._select_command

    def __eq__(self, other):
        """Command equality."""
        return (
            self.cmd_1 == other.cmd_1
            and self.cmd_2 == other.cmd_2
            and self.group_number == other.group_number
            and self.select == other.select
            and self.select_command == other.select_command
        )

    def __repr__(self):
        """String representation."""
        return "(cmd_1={}, cmd_2={}, gn={}, s={}, sc={})".format(
            binascii.hexlify(self.cmd_1).decode(),
            binascii.hexlify(self.cmd_2).decode(),
            self.group_number,
            self.select,
            self.select_command,
        )


class CommandSet:
    """Base class for command sets."""

    def __init__(
        self, group_number, brightness_steps, hue_steps=1, saturation_steps=1, temperature_steps=1
    ):
        """
        Initializes the command set.
        :param group_number: The group number.
        :param brightness_steps: The number of brightness steps.
        :param hue_steps: The number of hue steps.
        :param saturation_steps: The number of saturation steps
        :param temperature_steps: The number of temperature steps.
        """
        self._group_number = group_number
        self._brightness_steps = brightness_steps
        self._hue_steps = hue_steps
        self._saturation_steps = saturation_steps
        self._temperature_steps = temperature_steps

    @property
    def brightness_steps(self):
        """
        Brightness steps property.
        :return: The number of brightness steps.
        """
        return self._brightness_steps

    @property
    def hue_steps(self):
        """
        Color steps property.
        :return: The number of color steps.
        """
        return self._hue_steps

    @property
    def saturation_steps(self):
        """
        Saturation steps property.
        :return: The number of saturation steps.
        """
        return self._saturation_steps

    @property
    def temperature_steps(self):
        """
        Temperature steps property.
        :return: The number of temperature steps.
        """
        return self._temperature_steps



"""White LimitlessLED group."""

import time

from limitlessled import util
from limitlessled.group import Group, rate
from limitlessled.util import steps

WHITE = "white"


class WhiteGroup(Group):
    """White LimitlessLED group."""

    def __init__(self, bridge, number, name):
        """Initialize white group.

        Brightness and temperature must be initialized
        to some value.

        :param bridge: Associated bridge.
        :param number: Group number (1-4).
        :param name: Group name.
        """
        super().__init__(bridge, number, name, WHITE)
        self._temperature = 0.5

    def night_light(self):
        """Set night light mode."""
        cmd = self.command_set.night_light()
        self.send(cmd)

    @property
    def brightness(self):
        """Brightness property.

        :returns: Brightness.
        """
        return self._brightness

    @brightness.setter
    def brightness(self, brightness):
        """Set the brightness.

        :param brightness: Value to set (0.0-1.0).
        """
        try:
            cmd = self.command_set.brightness(brightness)
            self.send(cmd)
            self._brightness = brightness
        except AttributeError:
            self._setter(
                "_brightness", brightness, self._dimmest, self._brightest, self._to_brightness
            )

    @property
    def temperature(self):
        """Temperature property.

        :returns: Temperature (0.0-1.0)
        """
        return self._temperature

    @temperature.setter
    def temperature(self, temperature):
        """Set the temperature.

        :param temperature: Value to set (0.0-1.0).
        """
        try:
            cmd = self.command_set.temperature(temperature)
            self.send(cmd)
            self._temperature = temperature
        except AttributeError:
            self._setter(
                "_temperature", temperature, self._warmest, self._coolest, self._to_temperature
            )

    def transition(self, duration, brightness=None, temperature=None):
        """Transition wrapper.

        Short-circuit transition if necessary.

        :param duration: Duration of transition.
        :param brightness: Transition to this brightness.
        :param temperature: Transition to this temperature.
        """
        # Transition immediately if duration is zero.
        if duration == 0:
            if brightness is not None:
                self.brightness = brightness
            if temperature is not None:
                self.temperature = temperature
            return
        if brightness != self.brightness or temperature != self.temperature:
            self._transition(duration, brightness, temperature)

    @rate(reps=1)
    def _transition(self, duration, brightness, temperature):
        """Complete a transition.

        :param duration: Duration of transition.
        :param brightness: Transition to this brightness.
        :param temperature: Transition to this temperature.
        """
        # Set initial value.
        b_start = self.brightness
        t_start = self.temperature
        # Compute ideal step amount.
        b_steps = 0
        if brightness is not None:
            b_steps = steps(self.brightness, brightness, self.command_set.brightness_steps)
        t_steps = 0
        if temperature is not None:
            t_steps = steps(self.temperature, temperature, self.command_set.temperature_steps)
        # Compute ideal step amount (at least one).
        total_steps = max(b_steps, t_steps, 1)
        total_commands = b_steps + t_steps
        # Calculate wait.
        wait = self._wait(duration, total_steps, total_commands)
        # Scale down steps if no wait time.
        if wait == 0:
            b_steps, t_steps = self._scale_steps(duration, total_commands, b_steps, t_steps)
            total_steps = max(b_steps, t_steps, 1)
        # Perform transition.
        for i in range(total_steps):
            # Brightness.
            if b_steps > 0 and i % (total_steps / b_steps) == 0:
                self.brightness = util.transition(i, total_steps, b_start, brightness)
            # Temperature.
            elif t_steps > 0:
                self.temperature = util.transition(i, total_steps, t_start, temperature)
            # Wait.
            time.sleep(wait)

    def _setter(self, attr, value, bottom, top, to_step):
        """Set a value.

        :param attr: Attribute to set.
        :param value: Value to use.
        :param bottom: Get to bottom value.
        :param top: Get to top value.
        :param to_step: Get to intermediary value.
        """
        if value < 0 or value > 1:
            raise ValueError("out of range")
        if value == 0.0:
            bottom()
        elif value == 1.0:
            top()
        else:
            to_step(value)
        setattr(self, attr, value)

    def _to_brightness(self, brightness):
        """Step to a given brightness.

        :param brightness: Get to this brightness.
        """
        self._to_value(
            self._brightness,
            brightness,
            self.command_set.brightness_steps,
            self._dimmer,
            self._brighter,
        )

    def _to_temperature(self, temperature):
        """Step to a given temperature.

        :param temperature: Get to this temperature.
        """
        self._to_value(
            self._temperature,
            temperature,
            self.command_set.temperature_steps,
            self._warmer,
            self._cooler,
        )

    @rate(reps=1)
    def _to_value(self, current, target, max_steps, step_down, step_up):
        """Step to a value

        :param current: Current value.
        :param target: Target value.
        :param max_steps: Maximum number of steps.
        :param step_down: Down function.
        :param step_up: Up function.
        """
        for _ in range(steps(current, target, max_steps)):
            if (current - target) > 0:
                step_down()
            else:
                step_up()

    @rate(wait=0.025, reps=2)
    def _brightest(self):
        """Group as bright as possible."""
        for _ in range(steps(self.brightness, 1.0, self.command_set.brightness_steps)):
            self._brighter()

    @rate(wait=0.025, reps=2)
    def _dimmest(self):
        """Group brightness as dim as possible."""
        for _ in range(steps(self.brightness, 0.0, self.command_set.brightness_steps)):
            self._dimmer()

    @rate(wait=0.025, reps=2)
    def _warmest(self):
        """Group temperature as warm as possible."""
        for _ in range(steps(self.temperature, 0.0, self.command_set.temperature_steps)):
            self._warmer()

    @rate(wait=0.025, reps=2)
    def _coolest(self):
        """Group temperature as cool as possible."""
        for _ in range(steps(self.temperature, 1.0, self.command_set.temperature_steps)):
            self._cooler()

    def _brighter(self):
        """One step brighter."""
        self.send(self.command_set.brighter())

    def _dimmer(self):
        """One step dimmer."""
        self.send(self.command_set.dimmer())

    def _warmer(self):
        """One step warmer."""
        self.send(self.command_set.warmer())

    def _cooler(self):
        """One step cooler."""
        self.send(self.command_set.cooler())



"""LimitlessLED groups."""

import math
import queue
import threading
import time

from limitlessled import MIN_WAIT, REPS
from limitlessled.group.commands import command_set_factory
from limitlessled.pipeline import Pipeline, PipelineQueue


def rate(wait=MIN_WAIT, reps=REPS):
    """Rate limit a command function.

    :param wait: How long to wait between commands.
    :param reps: How many times to send a command.
    :returns: Decorator.
    """

    def decorator(function):
        """Decorator function.

        :returns: Wrapper.
        """

        def wrapper(self, *args, **kwargs):
            """Wrapper.

            :param args: Passthrough positional arguments.
            :param kwargs: Passthrough keyword arguments.
            """
            saved_wait = self.wait
            saved_reps = self.reps
            self.wait = wait
            self.reps = reps
            function(self, *args, **kwargs)
            self.wait = saved_wait
            self.reps = saved_reps

        return wrapper

    return decorator


class Group(object):
    """LimitlessLED group."""

    def __init__(self, bridge, number, name, led_type):
        """Initialize group.

        :param bridge: Member of this bridge.
        :param number: Group number (1-4).
        :param name: Group name.
        :param led_type: The type of the led.
        """
        self.name = name
        self.number = number
        self._bridge = bridge
        self._index = number - 1
        self._command_set = command_set_factory(bridge, number, led_type)
        self._on = False
        self._brightness = 0.5
        self._queue = queue.Queue()
        self._event = threading.Event()
        self._thread = PipelineQueue(self._queue, self._event)
        self._thread.daemon = True
        self._thread.start()
        self.wait = MIN_WAIT
        self.reps = REPS

    @property
    def on(self):
        """Is the group on?

        :return: True if the group is on, otherwise False.
        """
        return self._on

    @on.setter
    def on(self, state):
        """Turn on or off.

        :param state: True (on) or False (off).
        """
        self._on = state
        cmd = self.command_set.off()
        if state:
            cmd = self.command_set.on()
        self.send(cmd)

    def link(self):
        """Link new lights."""
        cmd = self.command_set.link()
        self.send(cmd)

    def unlink(self):
        """Unlink linked lights."""
        cmd = self.command_set.unlink()
        self.send(cmd)

    @property
    def bridge(self):
        """Bridge property."""
        return self._bridge

    @property
    def command_set(self):
        """Command set property."""
        return self._command_set

    def flash(self, duration=0.0):
        """Flash a group.

        :param duration: How quickly to flash (in seconds).
        """
        for _ in range(2):
            self.on = not self.on
            time.sleep(duration)

    def send(self, cmd):
        """Send a command to the bridge.

        :param cmd: List of command bytes.
        """
        self._bridge.send(cmd, wait=self.wait, reps=self.reps)

    def enqueue(self, pipeline):
        """Start a pipeline.

        :param pipeline: Start this pipeline.
        """
        copied = Pipeline().append(pipeline)
        copied.group = self
        self._queue.put(copied)

    def stop(self):
        """Stop a running pipeline."""
        self._event.set()

    def _wait(self, duration, steps, commands):
        """Compute wait time.

        :param duration: Total time (in seconds).
        :param steps: Number of steps.
        :param commands: Number of commands.
        :returns: Wait in seconds.
        """
        wait = ((duration - self.wait * self.reps * commands) / steps) - (
            self.wait * self.reps * self._bridge.active
        )
        return max(0, wait)

    def _scale_steps(self, duration, commands, *steps):
        """Scale steps

        :param duration: Total time (in seconds)
        :param commands: Number of commands to be executed.
        :param steps: Steps for one or many properties to take.
        :return: Steps scaled to time and total.
        """
        factor = duration / (
            (self.wait * self.reps * commands) - (self.wait * self.reps * self._bridge.active)
        )
        steps = [math.ceil(factor * step) for step in steps]
        if len(steps) == 1:
            return steps[0]
        else:
            return steps

    def __str__(self):
        """String representation.

        :returns: String
        """
        return "{} ({}) @ {}".format(self.number, self.name, self._bridge.ip)



"""Command sets for wifi bridge version 5 and lower."""

import math

from limitlessled.group.commands import Command, CommandSet
from limitlessled.group.rgbw import RGBW
from limitlessled.group.white import WHITE


class CommandLegacy(Command):
    """Represents a single v6 command to be sent to the bridge."""

    SUFFIX_BYTE = 0x00
    BRIDGE_SHORT_VERSION_MIN = 3
    BRIDGE_LONG_BYTE = 0x55

    def __init__(self, cmd_1, cmd_2, group_number, select=False, select_command=None):
        """
        Initialize command.
        :param cmd_1: The first part of the command.
        :param cmd_1: The second part of the command.
        :param group_number: Group number (1-4).
        :param select: If command requires selection.
        :param select_command: Selection command bytes.
        """
        super().__init__(cmd_1, cmd_2, group_number, select, select_command)

    def get_bytes(self, bridge):
        """
        Gets the full command as bytes.
        :param bridge: The bridge, to which the command should be sent.
        """
        if self.cmd_2 is not None:
            cmd = [self.cmd_1, self.cmd_2]
        else:
            cmd = [self.cmd_1, self.SUFFIX_BYTE]

        if bridge.version < self.BRIDGE_SHORT_VERSION_MIN:
            cmd.append(self.BRIDGE_LONG_BYTE)

        return bytearray(cmd)


class CommandSetLegacy(CommandSet):
    """Base command set for legacy wifi bridges."""

    SUPPORTED_VERSIONS = [1, 2, 3, 4, 5]
    BRIGHTNESS_OFFSET = 2

    def convert_brightness(self, brightness):
        """
        Convert the brightness from decimal percent (0.0-1.0)
        to byte representation for use in commands.
        :param brightness: The brightness from in decimal percent (0.0-1.0).
        :return: The brightness in byte representation.
        """
        brightness = math.ceil(brightness * self.brightness_steps)
        return brightness + self.BRIGHTNESS_OFFSET

    @staticmethod
    def convert_hue(hue):
        """
        Converts the hue from HSV color circle to the LimitlessLED color wheel.
        :param hue: The hue in decimal percent (0.0-1.0).
        :return: The hue regarding the LimitlessLED color wheel.
        """
        hue = hue * -1 + 1 + (2.0 / 3.0)  # RGB -> BGR

        return int(math.floor((hue % 1) * 256))

    def _build_command(self, cmd_1, cmd_2=None, select=False, select_command=None):
        """
        Constructs the complete command.
        :param cmd_1: Light command 1.
        :param cmd_2: Light command 2.
        :param select: If command requires selection.
        :param select_command: Selection command bytes.
        :return: The complete command.
        """

        return CommandLegacy(cmd_1, cmd_2, self._group_number, select, select_command)


class CommandSetWhiteLegacy(CommandSetLegacy):
    """Command set for white led light connected to legacy wifi bridge."""

    SUPPORTED_LED_TYPES = [WHITE]
    ON_BYTES = [0x38, 0x3D, 0x37, 0x32]
    OFF_BYTES = [0x3B, 0x33, 0x3A, 0x36]
    NIGHT_BYTES = [0xBB, 0xB3, 0xBA, 0xB6]
    BRIGHTNESS_STEPS = 10
    TEMPERATURE_STEPS = 10

    def __init__(self, group_number):
        """
        Initializes the command set.
        :param group_number: The group number.
        """
        super().__init__(
            group_number, self.BRIGHTNESS_STEPS, temperature_steps=self.TEMPERATURE_STEPS
        )

    def on(self):
        """
        Build command for turning the led on.
        :return: The command.
        """
        return self._build_command(self.ON_BYTES[self._group_number - 1])

    def off(self):
        """
        Build command for turning the led off.
        :return: The command.
        """
        return self._build_command(self.OFF_BYTES[self._group_number - 1])

    def night_light(self):
        """
        Build command for turning the led to night light mode.
        :return: The command.
        """
        return self._build_command(
            self.NIGHT_BYTES[self._group_number - 1], select=True, select_command=self.off()
        )

    def dimmer(self):
        """
        Build command for setting the brightness one step dimmer.
        :return: The command.
        """
        return self._build_command(0x34, select=True, select_command=self.on())

    def brighter(self):
        """
        Build command for setting the brightness one step brighter.
        :return: The command.
        """
        return self._build_command(0x3C, select=True, select_command=self.on())

    def cooler(self):
        """
        Build command for setting the temperature one step cooler.
        :return: The command.
        """
        return self._build_command(0x3F, select=True, select_command=self.on())

    def warmer(self):
        """
        Build command for setting the temperature one step warmer.
        :return: The command.
        """
        return self._build_command(0x3E, select=True, select_command=self.on())


class CommandSetRgbwLegacy(CommandSetLegacy):
    """Command set for RGBW led light connected to legacy wifi bridge."""

    SUPPORTED_LED_TYPES = [RGBW]
    HUE_STEPS = 255
    BRIGHTNESS_STEPS = 25

    def __init__(self, group_number):
        """
        Initializes the command set.
        :param group_number: The group number.
        """
        super().__init__(group_number, self.BRIGHTNESS_STEPS, hue_steps=self.HUE_STEPS)

    def on(self):
        """
        Build command for turning the led on.
        :return: The command.
        """
        return self._build_command(self._offset(0x45))

    def off(self):
        """
        Build command for turning the led off.
        :return: The command.
        """
        return self._build_command(self._offset(0x46))

    def night_light(self):
        """
        Build command for turning the led to night light mode.
        :return: The command.
        """
        return self._build_command(self._offset(0xC6), select=True, select_command=self.off())

    def white(self):
        """
        Build command for turning the led into white mode.
        :return: The command.
        """
        return self._build_command(self._offset(0xC5), select=True, select_command=self.on())

    def hue(self, hue):
        """
        Build command for setting the hue of the led.
        :param hue: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(
            0x40, self.convert_hue(hue), select=True, select_command=self.on()
        )

    def brightness(self, brightness):
        """
        Build command for setting the brightness of the led.
        :param brightness: Value to set (0.0-1.0).
        :return: The command.
        """
        return self._build_command(
            0x4E, self.convert_brightness(brightness), select=True, select_command=self.on()
        )

    def _offset(self, byte):
        """Calcuate group command offset.

        :param byte: Base byte.
        :returns: Appropriate byte for group.
        """
        index = self._group_number - 1
        return byte + (index * 2)



"""LimitlessLED pipeline."""

import logging
import threading
import time

from limitlessled import Color

_LOGGER = logging.getLogger(__name__)


class PipelineQueue(threading.Thread):
    """Pipeline queue."""

    def __init__(self, queue, event):
        """Initialize pipeline queue.

        :param queue: Read from this queue.
        :param event: Read from this event.
        """
        super(PipelineQueue, self).__init__()
        self._queue = queue
        self._event = event

    def run(self):
        """Run the pipeline queue.

        The pipeline queue will run forever.
        """
        while True:
            self._event.clear()
            self._queue.get().run(self._event)


class Stage(object):
    """Stage."""

    def __init__(self, name, args, kwargs):
        """Initialize stage.

        :param name: Name of stage.
        :param args: Positional arguments.
        :param kwargs: Keyword arguments.
        """
        self.name = name
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        """String representation of Stage.

        :returns: String
        """
        args = []
        if len(self.args) > 0:
            args += ["{}".format(a) for a in self.args]
        if len(self.kwargs) > 0:
            args += ["{}={}".format(k, v) for k, v in self.kwargs.items()]
        return "{}({})".format(self.name, ", ".join(args))


class Pipeline(object):
    """Pipeline.

    Pipelines are built by linking together a series
    of stages. Upon execution, the pipeline applies
    each stage to the given LimitlessLED group.
    """

    def __init__(self):
        """Initialize pipeline.

        Stage methods are dynamically initialized
        since they all follow the same pattern.
        """
        self._pipe = []
        self._group = None
        stages = [
            "on",
            "off",
            "color",
            "transition",
            "flash",
            "callback",
            "repeat",
            "brightness",
            "wait",
            "temperature",
            "white",
            "white_up",
            "white_down",
            "red_up",
            "red_down",
            "green_up",
            "green_down",
            "blue_up",
            "blue_down",
            "night_light",
            "link",
            "unlink",
        ]
        for name in stages:
            self._add_stage(name)

    @property
    def pipe(self):
        """Pipe property."""
        return self._pipe

    @property
    def group(self):
        """Group property."""
        return self._group

    @group.setter
    def group(self, group):
        """Set the associated group.

        :param group: Associated group.
        """
        self._group = group

    def run(self, stop):
        """Run the pipeline.

        :param stop: Stop event
        """
        _LOGGER.info("Starting a new pipeline on group %s", self._group)
        self._group.bridge.incr_active()
        for i, stage in enumerate(self._pipe):
            self._execute_stage(i, stage, stop)
        _LOGGER.info("Finished pipeline on group %s", self._group)
        self._group.bridge.decr_active()

    def append(self, pipeline):
        """Append a pipeline to this pipeline.

        :param pipeline: Pipeline to append.
        :returns: This pipeline.
        """
        for stage in pipeline.pipe:
            self._pipe.append(stage)
        return self

    def _add_stage(self, name):
        """Add stage methods at runtime.

        Stage methods all follow the same pattern.

        :param name: Stage name.
        """

        def stage_func(self, *args, **kwargs):
            """Stage function.

            :param args: Positional arguments.
            :param kwargs: Keyword arguments.
            :return: Pipeline (for method chaining).
            """
            self._pipe.append(Stage(name, args, kwargs))
            return self

        setattr(Pipeline, name, stage_func)

    def _execute_stage(self, index, stage, stop):
        """Execute a pipeline stage.

        :param index: Stage index.
        :param stage: Stage object.
        """
        if stop.is_set():
            _LOGGER.info("Stopped pipeline on group %s", self._group)
            return
        _LOGGER.info(" -> Running stage '%s' on group %s", stage, self._group)
        if stage.name == "on":
            self._group.on = True
        elif stage.name == "off":
            self._group.on = False
        elif stage.name == "hue":
            self._group.hue = stage.args[0]
        elif stage.name == "saturation":
            self._group.saturation = stage.args[0]
        elif stage.name == "color":
            self._group.color = Color(*stage.args)
        elif stage.name == "brightness":
            self._group.brightness = stage.args[0]
        elif stage.name == "temperature":
            self._group.temperature = stage.args[0]
        elif stage.name == "transition":
            self._group.transition(*stage.args, **stage.kwargs)
        elif stage.name == "white":
            self._group.white()
        elif stage.name == "white_up":
            self._group.white_up()
        elif stage.name == "white_down":
            self._group.white_down()
        elif stage.name == "red_up":
            self._group.red_up()
        elif stage.name == "red_down":
            self._group.red_down()
        elif stage.name == "green_up":
            self._group.green_up()
        elif stage.name == "green_down":
            self._group.green_down()
        elif stage.name == "blue_up":
            self._group.blue_up()
        elif stage.name == "blue_down":
            self._group.blue_down()
        elif stage.name == "night_light":
            self._group.night_light()
        elif stage.name == "link":
            self._group.link()
        elif stage.name == "unlink":
            self._group.unlink()
        elif stage.name == "flash":
            self._group.flash(**stage.kwargs)
        elif stage.name == "repeat":
            self._repeat(index, stage, stop)
        elif stage.name == "wait":
            time.sleep(*stage.args)
        elif stage.name == "callback":
            stage.args[0](*stage.args[1:], **stage.kwargs)

    def _repeat(self, index, stage, stop):
        """Repeat a stage.

        :param index: Stage index.
        :param stage: Stage object to repeat.
        :param iterations: Number of iterations (default infinite).
        :param stages: Stages back to repeat (default 1).
        """
        times = None
        if "iterations" in stage.kwargs:
            times = stage.kwargs["iterations"] - 1
        stages_back = 1
        if "stages" in stage.kwargs:
            stages_back = stage.kwargs["stages"]
        i = 0
        while i != times:
            if stop.is_set():
                break
            for forward in range(stages_back):
                if stop.is_set():
                    break
                stage_index = index - stages_back + forward
                self._execute_stage(stage_index, self._pipe[stage_index], stop)
            i += 1



"""Utility functions."""

from colorsys import hsv_to_rgb, rgb_to_hsv

from limitlessled import Color


def hue_of_color(color):
    """
    Gets the hue of a color.
    :param color: The RGB color tuple.
    :return: The hue of the color (0.0-1.0).
    """
    return rgb_to_hsv(*[x / 255 for x in color])[0]


def saturation_of_color(color):
    """
    Gets the saturation of a color.
    :param color: The RGB color tuple.
    :return: The saturation of the color (0.0-1.0).
    """
    return rgb_to_hsv(*[x / 255 for x in color])[1]


def to_rgb(hue, saturation):
    """
    Converts hue and saturation to RGB color.
    :param hue: The hue of the color.
    :param saturation: The saturation of the color.
    :return: The RGB color tuple.
    """
    return Color(*hsv_to_rgb(hue, saturation, 1))


def transition(value, maximum, start, end):
    """Transition between two values.

    :param value: Current iteration.
    :param maximum: Maximum number of iterations.
    :param start: Start value.
    :param end: End value.
    :returns: Transitional value.
    """
    return round(start + (end - start) * value / maximum, 2)


def steps(current, target, max_steps):
    """Steps between two values.

    :param current: Current value (0.0-1.0).
    :param target: Target value (0.0-1.0).
    :param max_steps: Maximum number of steps.
    """
    if current < 0 or current > 1.0:
        raise ValueError("current value %s is out of bounds (0.0-1.0)", current)
    if target < 0 or target > 1.0:
        raise ValueError("target value %s is out of bounds (0.0-1.0)", target)
    return int(abs((current * max_steps) - (target * max_steps)))



"""LimitlessLED Bridge."""

import queue
import select
import socket
import threading
import time

from limitlessled import MIN_WAIT, REPS
from limitlessled.group.dimmer import DIMMER, DimmerGroup
from limitlessled.group.rgbcct import RGBCCT, RgbcctGroup
from limitlessled.group.rgbw import BRIDGE_LED, RGBW, RgbwGroup
from limitlessled.group.rgbww import RGBWW, RgbwwGroup
from limitlessled.group.white import WHITE, WhiteGroup
from limitlessled.group.wrgb import WRGB, WrgbGroup

BRIDGE_PORT = 5987
BRIDGE_VERSION = 6
BRIDGE_LED_GROUP = 1
BRIDGE_LED_NAME = "bridge"
SELECT_WAIT = 0.025
BRIDGE_INITIALIZATION_COMMAND = [
    0x20,
    0x00,
    0x00,
    0x00,
    0x16,
    0x02,
    0x62,
    0x3A,
    0xD5,
    0xED,
    0xA3,
    0x01,
    0xAE,
    0x08,
    0x2D,
    0x46,
    0x61,
    0x41,
    0xA7,
    0xF6,
    0xDC,
    0xAF,
    0xFE,
    0xF7,
    0x00,
    0x00,
    0x1E,
]
KEEP_ALIVE_COMMAND_PREAMBLE = [0xD0, 0x00, 0x00, 0x00, 0x02]
KEEP_ALIVE_RESPONSE_PREAMBLE = [0xD8, 0x0, 0x0, 0x0, 0x07]
KEEP_ALIVE_TIME = 5
RECONNECT_TIME = 5
SOCKET_TIMEOUT = 5
STARTING_SEQUENTIAL_BYTE = 0x02


def group_factory(bridge, number, name, led_type):
    """Make a group.

    :param bridge: Member of this bridge.
    :param number: Group number (1-4).
    :param name: Name of group.
    :param led_type: Either `RGBW`, `WRGB`, `RGBWW`, `WHITE`, `DIMMER` or `BRIDGE_LED`.
    :returns: New group.
    """
    if led_type in [RGBW, BRIDGE_LED]:
        return RgbwGroup(bridge, number, name, led_type)
    elif led_type == RGBWW:
        return RgbwwGroup(bridge, number, name)
    elif led_type == WHITE:
        return WhiteGroup(bridge, number, name)
    elif led_type == DIMMER:
        return DimmerGroup(bridge, number, name)
    elif led_type == WRGB:
        return WrgbGroup(bridge, number, name)
    elif led_type == RGBCCT:
        return RgbcctGroup(bridge, number, name)
    else:
        raise ValueError("Invalid LED type: %s", led_type)


class Bridge(object):
    """Represents a LimitlessLED bridge."""

    def __init__(
        self, ip, port=BRIDGE_PORT, version=BRIDGE_VERSION, bridge_led_name=BRIDGE_LED_NAME
    ):
        """Initialize bridge.

        Bridge version 6 (latest as of this release)
        can use the default parameters. For lower versions,
        use port 8899 (3 to 5) or 50000 (lower then 3).
        Lower versions also require sending a larger payload
        to the bridge (slower).

        :param ip: IP address of bridge.
        :param port: Bridge port.
        :param version: Bridge version.
        :param bridge_led_name: Name of the bridge led group.
        """
        self.is_ready = True
        self.is_closed = False
        self.wait = MIN_WAIT
        self.reps = REPS
        self.groups = []
        self.ip = ip
        self.version = version
        self._sn = STARTING_SEQUENTIAL_BYTE
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.settimeout(SOCKET_TIMEOUT)
        self._socket.connect((ip, port))
        self._command_queue = queue.Queue()
        self._lock = threading.Lock()
        self.active = 0
        self._selected_number = None

        # Start queue consumer thread.
        self._consumer = threading.Thread(target=self._consume)
        self._consumer.daemon = True
        self._consumer.start()

        # Version specific stuff
        self._wb1 = None
        self._wb2 = None
        self._bridge_led = None
        if self.version >= 6:
            # Create bridge led group
            self._bridge_led = group_factory(self, BRIDGE_LED_GROUP, bridge_led_name, BRIDGE_LED)

            # Set is_ready to False before initing connection
            self.is_ready = False

            # Initialize connection to retrieve bridge session ids (wb1, wb2)
            self._init_connection()

            # Start keep alive thread.
            keep_alive_thread = threading.Thread(target=self._keep_alive)
            keep_alive_thread.daemon = True
            keep_alive_thread.start()

    @property
    def sn(self):
        """Gets the current sequential byte."""
        return self._sn

    @property
    def wb1(self):
        """Gets the bridge session id 1."""
        return self._wb1

    @property
    def wb2(self):
        """Gets the bridge session id 2."""
        return self._wb2

    @property
    def bridge_led(self):
        """Get the group to control the bridge led."""
        return self._bridge_led

    def incr_active(self):
        """Increment number of active groups."""
        with self._lock:
            self.active += 1

    def decr_active(self):
        """Decrement number of active groups."""
        with self._lock:
            self.active -= 1

    def add_group(self, number, name, led_type):
        """Add a group.

        :param number: Group number (1-4).
        :param name: Group name.
        :param led_type: Either `RGBW`, `WRGB`, `RGBWW`, `WHITE`, `DIMMER` or `BRIDGE_LED`.
        :returns: Added group.
        """
        group = group_factory(self, number, name, led_type)
        self.groups.append(group)
        return group

    def send(self, command, reps=REPS, wait=MIN_WAIT):
        """Send a command to the physical bridge.

        :param command: A Command instance.
        :param reps: Number of repetitions.
        :param wait: Wait time in seconds.
        """
        # Enqueue the command.
        self._command_queue.put((command, reps, wait))
        # Wait before accepting another command.
        # This keeps individual groups relatively synchronized.
        sleep = reps * wait * self.active
        if command.select and self._selected_number != command.group_number:
            sleep += SELECT_WAIT
        time.sleep(sleep)

    def _consume(self):
        """Consume commands from the queue.

        The command is repeated according to the configured value.
        Wait after each command is sent.

        The bridge socket is a shared resource. It must only
        be used by one thread at a time. Note that this can and
        will delay commands if multiple groups are attempting
        to communicate at the same time on the same bridge.
        """
        while not self.is_closed:
            # Get command from queue.
            msg = self._command_queue.get()

            # Closed
            if msg is None:
                return

            # Use the lock so we are sure is_ready is not changed during execution
            # and the socket is not in use
            with self._lock:
                # Check if bridge is ready
                if self.is_ready:
                    (command, reps, wait) = msg

                    # Select group if a different group is currently selected.
                    if command.select and self._selected_number != command.group_number:
                        if self._send_raw(command.select_command.get_bytes(self)):
                            self._selected_number = command.group_number
                            time.sleep(SELECT_WAIT)
                        else:
                            # Stop sending on socket error
                            self.is_ready = False

                    # Repeat command as necessary.
                    for _ in range(reps):
                        if self.is_ready:
                            if self._send_raw(command.get_bytes(self)):
                                time.sleep(wait)
                            else:
                                # Stop sending on socket error
                                self.is_ready = False

            # Wait if bridge is not ready, we're only reading is_ready, no lock needed
            if not self.is_ready and not self.is_closed:
                # For older bridges, always try again, there's no keep-alive thread
                if self.version < 6:
                    # Give the reconnect some time
                    time.sleep(RECONNECT_TIME)
                    self.is_ready = True

    def _send_raw(self, command):
        """
        Sends an raw command directly to the physical bridge.
        :param command: A bytearray.
        """
        try:
            self._socket.send(bytearray(command))
            self._sn = (self._sn + 1) % 256
            return True
        except (socket.error, socket.timeout):
            # We can get a socket.error or timeout exception if the bridge is disconnected,
            # but we are still sending data. In that case, return False to indicate that data is not sent.
            return False

    def _init_connection(self):
        """
        Requests the session ids of the bridge.
        :returns: True, if initialization was successful. False, otherwise.
        """
        try:
            # We are changing self.is_ready: lock it up!
            self._lock.acquire()

            response = bytearray(22)
            self._send_raw(BRIDGE_INITIALIZATION_COMMAND)
            self._socket.recv_into(response)
            self._wb1 = response[19]
            self._wb2 = response[20]
            self.is_ready = True
        except (socket.error, socket.timeout):
            # Connection timed out, bridge is not ready for us
            self.is_ready = False
        finally:
            # Prevent deadlocks: always release the lock
            self._lock.release()

        return self.is_ready

    def _reconnect(self):
        """
        Try continuously to reconnect to the bridge.
        """
        while not self.is_closed:
            if self._init_connection():
                return

            time.sleep(RECONNECT_TIME)

    def _keep_alive(self):
        """
        Send keep alive messages continuously to bridge.
        """
        send_next_keep_alive_at = 0
        while not self.is_closed:
            if not self.is_ready:
                self._reconnect()
                continue

            if time.monotonic() > send_next_keep_alive_at:
                command = KEEP_ALIVE_COMMAND_PREAMBLE + [self.wb1, self.wb2]
                self._send_raw(command)
                need_response_by = time.monotonic() + KEEP_ALIVE_TIME

            # Wait for responses
            timeout = max(0, need_response_by - time.monotonic())
            ready = select.select([self._socket], [], [], timeout)
            if ready[0]:
                try:
                    response = bytearray(12)
                    self._socket.recv_into(response)

                    if response[:5] == bytearray(KEEP_ALIVE_RESPONSE_PREAMBLE):
                        send_next_keep_alive_at = need_response_by
                except (socket.error, socket.timeout):
                    with self._lock:
                        self.is_ready = False
            elif send_next_keep_alive_at < need_response_by:
                # Acquire the lock to make sure we don't change self.is_ready
                # while _consume() is sending commands
                with self._lock:
                    self.is_ready = False

    def close(self):
        """
        Closes the connection to the bridge.
        """
        self.is_closed = True
        self.is_ready = False
        self._command_queue.put(None)

    def finish(self, timeout=None):
        self._command_queue.put(None)
        self._consumer.join(timeout)
        self.is_closed = True
        self.is_ready = False
